"""Default Morecantile TMS."""

import os
import pathlib
from copy import copy
from typing import Dict, List, Sequence, Union

import attr

from .errors import InvalidIdentifier
from .models import TileMatrixSet

morecantile_tms_dir = pathlib.Path(__file__).parent.joinpath("data")
tms_paths = list(pathlib.Path(morecantile_tms_dir).glob("*.json"))

user_tms_dir = os.environ.get("TILEMATRIXSET_DIRECTORY", None)
if user_tms_dir:
    tms_paths.extend(list(pathlib.Path(user_tms_dir).glob("*.json")))

default_tms: Dict[str, TileMatrixSet] = {
    tms.stem: TileMatrixSet.parse_file(tms) for tms in tms_paths
}


@attr.s(frozen=True)
class TileMatrixSets:
    """Default TileMatrixSets holder."""

    tms: Dict = attr.ib()

    def get(self, identifier: str) -> TileMatrixSet:
        """Fetch a TMS."""
        if identifier not in self.tms:
            raise InvalidIdentifier(f"Invalid identifier: {identifier}")

        return self.tms[identifier]

    def list(self) -> List[str]:
        """List registered TMS."""
        return list(self.tms.keys())

    def register(
        self,
        custom_tms: Union[TileMatrixSet, Sequence[TileMatrixSet]],
        overwrite: bool = False,
    ) -> "TileMatrixSets":
        """Register TileMatrixSet(s)."""
        if isinstance(custom_tms, TileMatrixSet):
            custom_tms = (custom_tms,)

        for tms in custom_tms:
            if tms.identifier in self.tms and not overwrite:
                raise Exception(f"{tms.identifier} is already a registered TMS.")

        new_tms = {tms.identifier: tms for tms in custom_tms}
        return TileMatrixSets({**self.tms, **new_tms})


tms = TileMatrixSets(copy(default_tms))  # noqa



"""Setup for morecantile."""

from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()

inst_reqs = ["attrs", "pyproj~=3.1", "pydantic"]

extra_reqs = {
    "rasterio": ["rasterio>=1.2.1"],
    "test": ["mercantile", "pytest", "pytest-cov"],
    "dev": ["pytest", "pytest-cov", "pre-commit"],
    "docs": ["mkdocs", "mkdocs-material", "pygments"],
}

setup(
    name="morecantile",
    version="3.0.2",
    python_requires=">=3.7",
    description=u"""Construct and use map tile grids (a.k.a TileMatrixSet / TMS).""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    keywords="GIS",
    author=u"Vincent Sarago",
    author_email="vincent@developmentseed.org",
    url="https://github.com/developmentseed/morecantile",
    license="MIT",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    package_data={"morecantile": ["data/*.json"]},
    zip_safe=False,
    install_requires=inst_reqs,
    extras_require=extra_reqs,
    entry_points="""
      [console_scripts]
      morecantile=morecantile.scripts.cli:cli
      """,
)



"""morecantile utils."""

import math
from typing import Dict, List, Tuple

from pyproj import CRS

from .commons import BoundingBox, Coords, Tile
from .errors import TileArgParsingError


def _parse_tile_arg(*args) -> Tile:
    """
    Parse the *tile arg of module functions

    Copy from https://github.com/mapbox/mercantile/blob/master/mercantile/__init__.py

    Parameters
    ----------
    tile : Tile or sequence of int
        May be be either an instance of Tile or 3 ints, X, Y, Z.

    Returns
    -------
    Tile

    Raises
    ------
    TileArgParsingError

    """
    if len(args) == 1:
        args = args[0]
    if len(args) == 3:
        return Tile(*args)
    else:
        raise TileArgParsingError(
            "the tile argument may have 1 or 3 values. Note that zoom is a keyword-only argument"
        )


def meters_per_unit(crs: CRS) -> float:
    """
    Coefficient to convert the coordinate reference system (CRS)
    units into meters (metersPerUnit).

    From note g in http://docs.opengeospatial.org/is/17-083r2/17-083r2.html#table_2:
        If the CRS uses meters as units of measure for the horizontal dimensions,
        then metersPerUnit=1; if it has degrees, then metersPerUnit=2pa/360
        (a is the Earth maximum radius of the ellipsoid).

    """
    # crs.linear_units_factor[1]  GDAL 3.0
    return (
        1.0 if crs.axis_info[0].unit_name == "metre" else 2 * math.pi * 6378137 / 360.0
    )


def truncate_lnglat(lng: float, lat: float) -> Tuple[float, float]:
    """
    Truncate Lat Lon Geographic coordinates.

    Copy from https://github.com/mapbox/mercantile/blob/master/mercantile/__init__.py

    """
    if lng > 180.0:
        lng = 180.0
    elif lng < -180.0:
        lng = -180.0

    if lat > 90.0:
        lat = 90.0
    elif lat < -90.0:
        lat = -90.0

    return lng, lat


def bbox_to_feature(west: float, south: float, east: float, north: float) -> Dict:
    """Create a GeoJSON feature from a bbox."""
    return {
        "type": "Polygon",
        "coordinates": [
            [[west, south], [west, north], [east, north], [east, south], [west, south]]
        ],
    }


def point_in_bbox(point: Coords, bbox: BoundingBox, precision: int = 5) -> bool:
    """Check if a point is in a bounding box."""
    return (
        round(point.x, precision) >= round(bbox.left, precision)
        and round(point.x, precision) <= round(bbox.right, precision)
        and round(point.y, precision) >= round(bbox.bottom, precision)
        and round(point.y, precision) <= round(bbox.top, precision)
    )


def is_power_of_two(number: int) -> bool:
    """Check if a number is a power of 2"""
    return (number & (number - 1) == 0) and number != 0


def check_quadkey_support(tms: List) -> bool:
    """Check if a Tile Matrix Set supports quadkeys"""
    return all(
        [
            (t.matrixWidth == t.matrixHeight)
            and is_power_of_two(t.matrixWidth)
            and ((t.matrixWidth * 2) == tms[i + 1].matrixWidth)
            for i, t in enumerate(tms[:-1])
        ]
    )
